# baseline options
from jf1uids.option_classes.simulation_config import SimulationConfig
from jf1uids.option_classes.simulation_params import SimulationParams

# module options
from jf1uids._physics_modules._stellar_wind.stellar_wind_options import WindParams
from jf1uids._physics_modules._stellar_wind.stellar_wind_options import WindConfig, MEO, MEI, EI